/* ExEf - the ultimate effect engine
Copyright (C) 2001-2003 Tom Olexa <olexat@post.cz> */

#include <errno.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <sys/mman.h>
#include <sched.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <signal.h>
#include <sys/file.h>

#include "ExEfDSP.h"

/* Global Variables */

#define BUF_SIZE 1024 /* fragment length in samples */
#define X 4
#define Y 10

    int Audio_fd;	/* variables for initializing soundcard */
    signed short Audio_buffer[BUF_SIZE];
    long Audio_buffer_32[BUF_SIZE];
    char Audio_DEVICE_NAME[40] = "/dev/dsp";

    int Audio_caps;
    int Audio_format = AFMT_S16_LE;
    int Audio_stereo = 0;
    int Audio_speed = 44100;
    int Audio_arg = 0x0004000a;
    int Audio_maxfrag;
    int Audio_24 = 0;
    
    audio_buf_info Audio_info;

    short Audio_BufSize = 128;	/* buffers */

    float xn [4] [BUF_SIZE];
    float xno [BUF_SIZE];


    FILE *Tcl_strmw;	/* Tcl channels */
    int Tcl_filer;
    char Tcl_line[90];
    int Tcl_charpos = 0;
    
    int Tcl_interactive = 0;

    int VU_nth = 1;		/* VU meters */
    float VU_inpower = 0, VU_interpower = 0, VU_outpower = 0;	
    float VU_lambda = 0.99;
    short VU_refresh = 50;
    int VU_onoff = 1;
    int VU_over, VU_inover = 0, VU_interover = 0, VU_outover = 0;

    short Switch_channel = 0;		/* output switch */
    short Switch_prevchan = 0;
    float Switch_lambda = 0.99;
    float Switch_gain = 0;
    short Switch_reset = 0;
    
    int Save_saving = 0;	/* Saving Audio */
    int Save_point;
    int Save_maxfrags = 8;
    int Save_fragwpage = 0;
    int Save_pages;
    char Save_name[80];

    SavedPage *start = NULL;
    SavedPage *current;

    int input_fd, input_src = 0, output_dest = 0;	/* IO redirection */
     
    
    Box Uarray [Y+1] [X];

    Gain *TGain;
    Pass *TPass;
    Biquad *TBiquad;
    Distort *TDistort;
    Half *THalf;
    Modulator *TModulator;
    Delay *TDelay;
    Hall *THall;
    Reverb *TReverb;
    Compressor *TCompressor;
    Quant *TQuant;
    NoiseGate *TNoiseGate;
    SpecSubF *TSpecSubF;
    SinGen *TSinGen;
    WNoise *TWNoise;


/*Design functions*/

float dB(short slide)
{
    return(powf(10,(float)slide/20));
}

float Forget(float tau)
{
    float beta;
    beta = (float)(Audio_speed*tau)/1000;
    return(1-1/beta);
}

unsigned short Distance(short l)
{
    float d;
    d = (float)(l*Audio_speed)/330;
    return((short)d);
}

void DesignPass(char ftion, short freq, float sg)
{
    float wp,wz,T,t1,t2,den;

    T = 1 / (float) Audio_speed;
    wp = 2*Audio_speed*tanf(M_PI*freq*T);
    t2 = 1/wp;
    den = (T + 2*t2);

    TPass->a1 = (T-2*t2)/den;

    if (ftion == 'l') {
	wz = wp/sg;
    } else {
	wz = wp*sg;
	den = den/sg;
    }

    t1 = 1/wz;
    
    TPass->b0 = (T+2*t1)/den;
    TPass->b1 = (T-2*t1)/den;

    printf("a1 = %e, b0 = %e, b1 = %e\n",TPass->a1,TPass->b0,TPass->b1);
}

void DesignBiquad(char ftion, short freq, float Q, float sg)
{
    float T,w,wz,den,QwT,bwT;

    T = 1 / (float) Audio_speed;
    if (ftion != 'e') Q = Q/(1+sg);
    w = 2*Audio_speed*tanf(M_PI*freq*T);
    QwT = w*w*Q*T*T;
    den = 4*Q + 2*w*T + QwT;

    TBiquad->a1 = (-8*Q + 2*QwT)/den;
    TBiquad->a2 = (4*Q - 2*w*T + QwT)/den;

    if (ftion == 'e') {
	TBiquad->b0 = (4*Q + 2*sg*w*T + QwT)/den;
	TBiquad->b1 = TBiquad->a1;
	TBiquad->b2 = (4*Q - 2*sg*w*T + QwT)/den;
    } else {
	if (ftion == 'h') {
    	    den = 0.5*den/Q;
            wz = w*sqrtf(sg);
	} else {
	    den = 0.5*den/Q/sg;
	    wz = w/sqrtf(sg);
	}
	bwT = 0.5*wz*wz*T*T;
	TBiquad->b0 = (2 + 2*wz*T + bwT)/den;
	TBiquad->b1 = (-4 + 2*bwT)/den;
	TBiquad->b2 = (2 - 2*wz*T + bwT)/den;
    }

    printf("a1 = %e, a2 = %e,",TBiquad->a1,TBiquad->a2);
    printf(" b0 = %e, b1 = %e, b2 = %e\n",TBiquad->b0,TBiquad->b1,TBiquad->b2);
}

void DesignReverb(short ord, unsigned short implen, unsigned short pre, unsigned short tau)
{
    short i;
    float r;
    for (i=0;i<ord;i++) {
        r = ((float) rand())/RAND_MAX;
	TReverb->Bi[i] = r*implen+pre;
	TReverb->Bv[i] = (2*(float) rand() - 1)/RAND_MAX*exp(-r*implen/tau);
    }
}


/*Terminating Functions*/

void Signals()
{
    exit(0);
}

void Errors(int errnum, int killnow, char *path)
{
    pid_t ppid = 0;
    fprintf(stderr,"This error if fatal, quitting");
    if (Tcl_interactive) {
	ppid = getppid();
	if (errnum != 0) {
	    if (path == NULL)
		fprintf(Tcl_strmw,"Error %d X\n", errnum);
	    else
		fprintf(Tcl_strmw,"Error %d %s\n", errnum, path);
	}
	if (killnow == 1) {
	    fprintf(stderr,"\nKilling ExEf GUI, pid %d\n",ppid);
	    kill(ppid, SIGTERM);
	} else {
	    fprintf(stderr," in 5 sec\n");
	    sleep(5);
    	    fprintf(stderr,"Killing ExEf GUI, pid %d\n",ppid);
    	    kill(ppid, SIGTERM);
	}
    } else fprintf(stderr, "\n");
    exit(1);
}

void SaveToDisk()
{
    int Saved, pagesize, n;
    SavedPage *temp;
    struct {
	char RIFF[4]; /* "RIFF" */
	long RiffLen; /* remaining bytes in File */
	char WaveFmt[8]; /* "WAVEfmt " */
	long FmtLen; /* 16, length of format block following */
	short FormatTag; /* 1 (WAVE_FORMAT_PCM), others reserved */
	short Channels; /* 1 */
	long SamplesPerSecond; /* in Hz (11025, 22050,44100,..) */
	long AvgBytesPerSec; /* (dwSamplesPerSecond * wBlockAlign) */
	short BlockAlign; /* (wBitsPerSample * wChannels / 8) */
	short BitsPerSample; /* 16 or 32 */
	char Data[4]; /* "data" */
	long DataLen; /* length of sample data (bytes!) */
	} WaveHeader;

    int Bps;

    if (Audio_24)
	Bps = 4;
    else
	Bps = 2;
    
    if (Save_saving) {
	printf("Saving file to disk\n");

	strncpy(WaveHeader.RIFF,"RIFF",4);
	WaveHeader.RiffLen = Bps*Save_pages*Audio_BufSize*Save_maxfrags + 36;
	strncpy(WaveHeader.WaveFmt,"WAVEfmt ",8);
	WaveHeader.FmtLen = 16;
	WaveHeader.FormatTag = 1;
	WaveHeader.Channels = 1;
	WaveHeader.SamplesPerSecond = Audio_speed;
	WaveHeader.AvgBytesPerSec = Audio_speed*Bps;
	WaveHeader.BlockAlign = Bps;
	WaveHeader.BitsPerSample = 8*Bps;
	strncpy(WaveHeader.Data,"data",4);
	WaveHeader.DataLen = WaveHeader.RiffLen - 36;

	if (Audio_24)
	    for (n=Save_fragwpage*Audio_BufSize;n<Save_maxfrags*Audio_BufSize;n++) {
	        current->data_32[n] = 0;
	    }
	else
	    for (n=Save_fragwpage*Audio_BufSize;n<Save_maxfrags*Audio_BufSize;n++) {
	        current->data[n] = 0;
	    }

	pagesize = Bps*Audio_BufSize*Save_maxfrags;
	current->next = NULL;
	current = start;
	if ((Saved = creat(Save_name,0644)) == -1) {
	    perror("Can't create file");
	    return;
	    if (Tcl_interactive) {
		fprintf(Tcl_strmw,"Error 8 %s\n", Save_name);
	    }
	}
	printf("Saving %lu bytes into %s...\n", WaveHeader.RiffLen + 8, Save_name);
	write(Saved, &WaveHeader, 44);
	while (current) {
	    write(Saved, current->data, pagesize);
	    current = current->next;
	}
	printf("Done\nFreeing memory...\n");
	current = start;
	while (current) {
	    temp = current->next;
	    free(current);
	    current = temp;
	}
	printf("Done\n");
    }
}

void InEof()
{
    if (input_src) {
	printf("End of file\n");
	if (output_dest)
	    exit(0);
	input_src = 0;
	close(Audio_fd);
	realtime();
	opencard();
	checkout();
	printf("fragstotal %d\n",Audio_info.fragstotal);
	printf("fragsize %d B\n",Audio_info.fragsize);
	Audio_maxfrag = Audio_info.fragstotal;
	input_fd = Audio_fd;
    }
}


/*Initialization functions - Data Structures*/

void CmdLine(int argc, char **argv, char **file, char **infile)
{
    char cwd[80];
    while (argc > 1) {
	argc--;
	argv++;

        if (**argv == '-') {
	    switch ((*argv)[1]) {
	        case 'i' :
		    printf("Interactive mode\n");
		    getcwd(cwd, 80);
	    	    Tcl_interactive = 1;
		    IniTmp();
		    OpenTcl();
		    chdir(cwd);
		    break;
	        case 'f' :
		    argc--;
		    if (argc == 0)
			{
			fprintf(stderr,"%s: filename missing\n", *argv);
			Errors(20, 0, NULL);
			}
		    argv++;
		    *infile = *argv;
		    break;
	    	case 'q' :
		    printf("Quiet operation\n");
		    output_dest = 1;
		    break;
	        default :
		    printf("Usage: ExEf [<options>] [config file]\n");
		    printf(" options are:\n"); 
	            printf(" [ -f file ] 	wav/raw file as input\n");
		    printf(" [ -q ]	 	quiet fast processing, quits at end of file\n\n");
		    Errors(21, 0, NULL);
	    }
	}
	else *file = *argv;
    }
}

void Reset(float *start)
{
    int i;
    for (i=0;i<65536;i++)
	start[i] = 0.0;
}

void ClrUarray()
{
    short x,y;
    for (x=0;x<X;x++) {
	for (y=0;y<Y+1;y++)
	    (Uarray [y] [x]).LinkToBox=NULL;
    }
}

void CreateBox(short x,short y,char kind)
{
    int i = 0;
    (Uarray [y] [x]).KindOfBox=kind;

    do {
	switch(kind) {
	    case 'g':
		(Uarray [y] [x]).LinkToBox=calloc(1,sizeof(Gain));
	    break;
	    case 'p':
		(Uarray [y] [x]).LinkToBox=calloc(1,sizeof(Pass));
		break;
	    case 'b':
		(Uarray [y] [x]).LinkToBox=calloc(1,sizeof(Biquad));
		break;
	    case 'd':
		(Uarray [y] [x]).LinkToBox=calloc(1,sizeof(Distort));
		break;
	    case 'h':
		(Uarray [y] [x]).LinkToBox=calloc(1,sizeof(Half));
		break;
	    case 'm':
		(Uarray [y] [x]).LinkToBox=calloc(1,sizeof(Modulator));
		break;
	    case 'D':
		(Uarray [y] [x]).LinkToBox=calloc(1,sizeof(Delay));
		break;
	    case 'H':
		(Uarray [y] [x]).LinkToBox=calloc(1,sizeof(Hall));
		break;
	    case 'R':
		(Uarray [y] [x]).LinkToBox=calloc(1,sizeof(Reverb));
		break;
	    case 'C':
		(Uarray [y] [x]).LinkToBox=calloc(1,sizeof(Compressor));
		break;
	    case 'q':
		(Uarray [y] [x]).LinkToBox=calloc(1,sizeof(Quant));
		break;
	    case 'n':
		(Uarray [y] [x]).LinkToBox=calloc(1,sizeof(NoiseGate));
		break;
	    case 'S':
		(Uarray [y] [x]).LinkToBox=calloc(1,sizeof(SpecSubF));
		break;
	    case 's':
		(Uarray [y] [x]).LinkToBox=calloc(1,sizeof(SinGen));
		break;
	    case 'w':
		(Uarray [y] [x]).LinkToBox=calloc(1,sizeof(WNoise));
		break;
	    case 'c':
		(Uarray [y] [x]).LinkToBox=(void *)!NULL;
		break;
	    case 'v':
		(Uarray [y] [x]).LinkToBox=(void *)!NULL;
		break;
	}
	i++;
    } while (((Uarray [y] [x]).LinkToBox == NULL) && (i < 3));
    if ((Uarray [y] [x]).LinkToBox == NULL) {
	fprintf(stderr, "Memory allocation failed\n");
	Errors(22, 0, NULL);
    }
}

void LoadBox(char *str,short x,short y,char kind)
{
    char ftion,ftion2;
    short slide,slide2,slide3,slide4,freq,tau,tau2,x1,x2,x3,pre,ord,reset = 0;
    unsigned short levels;

    switch(kind) {
        case 'g':
	    TGain = (Uarray [y] [x]).LinkToBox;
	    sscanf(str,"%hd",&slide);
	    printf("Gain %hd dB\n",slide);
	    TGain->g = dB(slide);
	    break;
	case 'p':
	    TPass = (Uarray [y] [x]).LinkToBox;
	    sscanf(str," %c %hd %hd",&ftion,&freq,&slide);
	    printf("Pass type %c, ",ftion);
	    printf("Frequency %hd Hz, ",freq);
	    printf("Stopband gain %hd dB\n",slide);
	    DesignPass(ftion,freq,dB(slide));
	    break;
	case 'b':
	    TBiquad = (Uarray [y] [x]).LinkToBox;
            sscanf(str," %c %hd %hd %hd",&ftion,&freq,&slide,&slide2);
	    printf("Pass type %c, ",ftion);
	    printf("Frequency %hd Hz, ",freq);
	    printf("Q factor %hd dB, ",slide);
            printf("Stopband gain %hd dB\n",slide2);
	    DesignBiquad(ftion,freq,dB(slide),dB(slide2));
	    break;
	case 'd':
	    TDistort = (Uarray [y] [x]).LinkToBox;
	    sscanf(str," %c %c %hd",&ftion,&ftion2,&slide);
	    printf("Left distortion %c, ",ftion);
	    printf("Right distortion %c, ",ftion2);
	    printf("ParaDir %hd dB\n",slide);
	    TDistort->lftion = ftion;
	    TDistort->rftion = ftion2;
	    TDistort->pd = dB(slide);
	    break;
	case 'h':
	    THalf = (Uarray [y] [x]).LinkToBox;
	    sscanf(str,"%hd",&slide);
	    printf("Paradir %hd dB\n",slide);
	    THalf->pd = dB(slide);
	    break;
	case 'm':
	    TModulator = (Uarray [y] [x]).LinkToBox;
	    sscanf(str," %c %c %hd %hd %hd",&ftion,&ftion2,&slide,&tau,&slide2);
	    printf("Modulator type %c, ",ftion);
	    printf("Wave form %c, ",ftion2);
	    printf("Modulation depth %hd, ",slide);
	    printf("Period %hd ms, ",tau);
	    printf("ParaDir %hd dB\n",slide2);
	    TModulator->ftion = ftion;
	    TModulator->wave = ftion2;
	    TModulator->depth = slide;
	    TModulator->period = tau;
	    TModulator->pd = dB(slide2);
	    break;
	case 'D':
	    TDelay = (Uarray [y] [x]).LinkToBox;
	    sscanf(str,"%hd %hd %hd %hd %hd %hd %hd",&pre,&x1,&x2,&slide,&slide2,&slide3,&reset);
	    printf("Predelay %hd m, ",pre);
	    printf("X1 %hd m, ",x1);
	    printf("X2 %hd m, ",x2);
	    printf("R1 %hd dB, ",slide);
	    printf("R2 %hd dB, ",slide2);
	    printf("Output %hd dB, ",slide3);
	    printf("Reset %hd\n",reset);
	    TDelay->pre = Distance(pre);
	    TDelay->d1 = 2*Distance(x1);
	    TDelay->d2 = 2*Distance(x2);
	    TDelay->R1 = dB(slide);
	    TDelay->R2 = dB(slide2);
	    TDelay->output = dB(slide3);
	    TDelay->reset = reset;
	    break;
	case 'H':
	    THall = (Uarray [y] [x]).LinkToBox;
	    sscanf(str,"%hd %hd %hd %hd %hd %hd %hd %hd %hd",&pre,&x1,&x2,&x3,&slide,&slide2,&slide3,&slide4,&reset);
	    printf("Predelay %hd m, ",pre);
	    printf("X %hd m, ",x1);
	    printf("Y %hd m, ",x2);
	    printf("Z %hd m, ",x3);
	    printf("Rx %hd dB, ",slide);
	    printf("Ry %hd dB, ",slide2);
	    printf("Rz %hd dB, ",slide3);
	    printf("Output %hd dB, ",slide4);
	    printf("Reset %hd\n",reset);
	    THall->pre = Distance(pre);
	    THall->d1 = 2*Distance(x1);
	    THall->d2 = 2*Distance(x2);
	    THall->d3 = 2*Distance(x3);
	    THall->R1 = dB(slide);
	    THall->R2 = dB(slide2);	    
	    THall->R3 = dB(slide3);
	    THall->output = dB(slide4);
	    THall->reset = reset;
	    break;
	case 'R':
	    TReverb = (Uarray [y] [x]).LinkToBox;
	    sscanf(str,"%hd %hd %hd %hd %hd %hd %hd %hd",&ord,&x1,&pre,&tau,&x2,&slide,&slide2,&reset);
	    printf("Order %hd, ",ord);
	    printf("Filter length %hd m, ",x1);
	    printf("Predelay %hd m, ",pre);
	    printf("Decay time constant %hd m, ",tau);
	    printf("Feedback time %hd m, ",x2);
            printf("Feedback %hd dB, ",slide);
            printf("Output %hd dB, ",slide2);
	    printf("Reset %hd\n",reset);
	    TReverb->ord = ord;
	    DesignReverb(ord,2*Distance(x1),Distance(pre),2*Distance(tau));
	    TReverb->fbtime = 2*Distance(x2);
	    TReverb->fbk = dB(slide);
	    TReverb->output = dB(slide2);
	    TReverb->reset = reset;
	    break;
	case 'C':
	    TCompressor = (Uarray [y] [x]).LinkToBox;
	    sscanf(str,"%hd %hd %hd %hd",&slide,&slide2,&tau,&tau2);
	    printf("Threshold %hd dB, ",slide);
	    printf("Compression ratio %hd dB / 10 dB, ",slide2);
	    printf("Attack time constant %hd ms, ",tau);
	    printf("Decay time constant %hd ms\n",tau2);
	    TCompressor->thresh = dB(slide - 3);
	    TCompressor->ratio = 0.1*slide2;
	    TCompressor->clambda = Forget(tau);
	    TCompressor->dlambda = Forget(tau2);
	    if (tau < tau2)
		TCompressor->lambda = TCompressor->clambda;
	    else
		TCompressor->lambda = TCompressor->dlambda;
	    break;
	case 'q':
	    TQuant = (Uarray [y] [x]).LinkToBox;
	    sscanf(str,"%hd",&slide);
	    printf("Max SNR reduced by %hd dB, ",slide);
	    levels = powf(2,(float)(96 - slide)/6);
	    printf("using %hd levels\n",levels);
	    TQuant->levels = 0.5*(float)levels;
	    break;
        case 'n':
            TNoiseGate = (Uarray [y] [x]).LinkToBox;
	    sscanf(str,"%hd %hd %hd %hd",&slide,&slide2,&tau,&tau2);
	    printf("Threshold %hd dB, ",slide);
	    printf("Power saturation %hd dB, ",slide2);
	    printf("Attack time constant %hd ms, ",tau);
	    printf("Decay time constant %hd ms\n",tau2);
	    TNoiseGate->threshp = 0.5*dB(slide)*dB(slide);
	    TNoiseGate->saturp = dB(slide2)*dB(slide2);
	    TNoiseGate->clambda = Forget(tau);
	    TNoiseGate->dlambda = Forget(tau2);
	    if (tau < tau2)
		TNoiseGate->lambda = Forget(tau);
	    else
		TNoiseGate->lambda = Forget(tau2);
	    break;
	case 'S':
	    TSpecSubF = (Uarray [y] [x]).LinkToBox;
	    sscanf(str,"%hd %hd %hd",&slide,&tau,&tau2);
	    printf("Threshold %hd dB, ",slide);
	    printf("Noise estimation time %hd ms, ",tau);
	    printf("Estimation end %hd ms\n",tau2);
	    TSpecSubF->thresh = 0.5*dB(slide)*dB(slide);
	    TSpecSubF->lambda = Forget((float)(tau)/Audio_BufSize);
	    TSpecSubF->limit = tau2;
	    TSpecSubF->time = 0;
	    break;
	case 's':
	    TSinGen = (Uarray [y] [x]).LinkToBox;
	    sscanf(str,"%hd %hd",&freq,&slide);
	    printf("Frequency %hd Hz, ",freq);
	    printf("ParaDir %hd dB\n",slide);
	    TSinGen->freq = freq;
	    TSinGen->pd = dB(slide);
	    break;
	case 'w':
	    TWNoise = (Uarray [y] [x]).LinkToBox;
	    sscanf(str,"%hd",&slide);
	    printf("ParaDir %hd dB\n",slide);
	    TWNoise->pd = dB(slide);
	    break;
	case 'c':
	    printf("No params\n");
	    break;
	case 'v':
	    printf("No params\n");
	    break;
    }
}

void ChangeBox(short x, short y, char kind)
{
    char s[40];
    char t;
    t = (Uarray [y] [x]).KindOfBox;
    if (( t != 'c') && (t != 'v')) {
        free((Uarray [y] [x]).LinkToBox);
    }
    CreateBox(x,y,kind);

    switch(kind) {
        case 'g':
	    strcpy(s,"0");
	    break;
	case 'p':
	    strcpy(s,"l 1000 -20");
	    break;
	case 'b':
	    strcpy(s,"l 1000 -6 -20");
	    break;
	case 'd':
	    strcpy(s,"a a -100");
	    break;
	case 'h':
	    strcpy(s,"-20");
	    break;
	case 'm':
	    strcpy(s,"a s 30 1000 -10");
	    break;
	case 'D':
	    strcpy(s,"5 10 15 -5 -10 -5");
	    break;
	case 'H':
	    strcpy(s,"5 15 25 35 -15 -20 -25 0");
	    break;
	case 'R':
	    strcpy(s,"20 20 5 10 20 -10 -10");
	    break;
	case 'C':
	    strcpy(s,"-40 5 1 10");
	    break;
	case 'q':
	    strcpy(s,"12");
	    break;
	case 'n':
	    strcpy(s,"-75 5 1 10");
	    break;
	case 'S':
	    strcpy(s,"-75 100 -1");
	    break;
	case 's':
	    strcpy(s,"1000 -20");
	    break;
	case 'w':
	    strcpy(s,"-20");
	    break;
	case 'c':
	    strcpy(s,"");
	    break;
	case 'v':
	    strcpy(s,"");
	    break;
    }

    fprintf(Tcl_strmw,"Box %d %d %c %s\n",x,y,kind,s);
	
    printf("Effect type %c, ",kind);
    LoadBox(s,x,y,kind);
}    

void InsertBox(short x, short y)
{
    short i;
    for (i = 9; i>y; i--) {
	(Uarray [i] [x]).KindOfBox = (Uarray [i - 1] [x]).KindOfBox;
	(Uarray [i] [x]).LinkToBox = (Uarray [i - 1] [x]).LinkToBox;
    }
    CreateBox(x,y,'g');	
    fprintf(Tcl_strmw,"Box %d %d g 0\n",x,y);
    LoadBox("0",x,y,'g');
}

void RemoveBox(short x, short y)
{
    short i;
    char t;
    t = (Uarray [y] [x]).KindOfBox;
    if (( t != 'c') && (t != 'v')) {
        free((Uarray [y] [x]).LinkToBox);
    }
    i = y;
    while ((Uarray [i] [x]).LinkToBox) {
	(Uarray [i] [x]).KindOfBox = (Uarray [i + 1] [x]).KindOfBox;
	(Uarray [i] [x]).LinkToBox = (Uarray [i + 1] [x]).LinkToBox;
	i++;
    }
}

void RemoveColumn(short x)
{
    short y;
    char t;

    printf("Removing column %hd\n",x);
    y = 0;
    while ((Uarray [y] [x]).LinkToBox) {
	t = (Uarray [y] [x]).KindOfBox;
	if (( t != 'c') && (t != 'v'))
    	    {
	    free((Uarray [y] [x]).LinkToBox);
    	    }
	(Uarray [y] [x]).LinkToBox = NULL;
	y++;
    }
}

void RemoveAll()
{
    short x;
    printf("Removing all boxes\n");
    for (x = 0; x < 4; x++)
	RemoveColumn(x);
}

void HomeDir(char *dir)
{
    char *home;
    char defaultPath[80];
    struct stat buf;

    home = getenv("HOME");
    if (home != NULL) {
    	sprintf(defaultPath, "%s/%s", home, ".ExEf");
    }

    printf("Checking %s directory\n",defaultPath);
    if (stat(defaultPath, &buf) == -1) {
	perror("Can't access");
	printf("Copying config files from /usr/share/ExEf\n");
	system("cp -v -R /usr/share/ExEf ~");
	system("mv -v -f ~/ExEf ~/.ExEf");
	system("rm -v -f ~/.ExEf/ExEflogo.gif");
    }
}

void ReadColumn(FILE *strm, short x)
{    
    char s[40];
    char a;
    short y=0;
    
    do {
	fscanf(strm," %c",&a);
	if (a != '*') {
	    printf("x=%d   y=%d\n",x,y);
	    printf("Effect type %c, ",a);
	    CreateBox(x,y,a);
	    fgets(s,40,strm);
	    if (Tcl_interactive) {
		fprintf(Tcl_strmw,"Box %d %d %c%s",x,y,a,s);
	    }
	    LoadBox(s,x,y,a);
	    y++;
	}    
    } while (a != '*');
}	
        
int ReadConfig(char *path)
{
    FILE *strm;
    
    char s[40];
    char a;
    short x,d;
    
    ClrUarray();

    printf("Opening %s config file\n",path);
    if ((strm=fopen(path,"r")) == NULL) {
	perror("Can't read config file");
	if (Tcl_interactive) {
	    fprintf(Tcl_strmw,"Error 1 %s\n",path);
	}
	return 0;
    }
    fgets(s,40,strm);
    if (strcmp(s,"ExEf config\n")) {
	printf("Config file broken\n");
	if (Tcl_interactive) {
	    fprintf(Tcl_strmw,"Error 2 %s\n",path);
	}
	fclose(strm);
	return 0;
    }

    fgets(s,40,strm);
    while (strcmp(s,"*Effects\n")) {
	printf(s);
	if (Tcl_interactive) {
	    fprintf(Tcl_strmw,s);
	}
	if (!strncmp(s,"Audio",5)) {	
	    d = 0;
	    sscanf(s + 6,"%d %hd %hd",&Audio_speed,&x,&d);
	    Audio_arg = 0x00040000 + x;
	    Audio_BufSize = 1<<(x-1);
	    Audio_24 = (d>16) ? 1:0;
	}
	if (!strncmp(s,"Device",6)) {
	    sscanf(s + 7,"%s",Audio_DEVICE_NAME);
	}
	if (!strncmp(s,"Switch",6)) {	
	    sscanf(s + 7,"%hd",&Switch_channel);
	    Switch_prevchan = Switch_channel;
	    sscanf(s + 11,"%hd",&x);
	    Switch_lambda = Forget(x);	
	}
	if (!strncmp(s,"VUMeters",8)) {
	    sscanf(s + 9,"%hd %hd %d",&x,&VU_refresh,&VU_onoff);
	    VU_lambda = Forget(x);
	}
	if (!strncmp(s,"SaveWave",8)) {
	    sscanf(s + 9,"%d %hd %s %s",&Save_saving,&x,Save_name,s);
	    if (!strcmp(s,"input"))
		Save_point = 1;
	    else Save_point = 2;
		Save_maxfrags = 1<<x;
	}
	fgets(s,40,strm);
    }

    x=0;

    fscanf(strm," %c",&a);
    fgets(s,40,strm);
    while (strncmp(s,"Column",6) == 0) {
	printf("Loading Column %d:%s",x,s+6);
	if (Tcl_interactive) {
	    fprintf(Tcl_strmw,"Column %d%s",x,s+6);
	}

	ReadColumn(strm,x);

	x++;
        fgets(s,40,strm);
    }

    fclose(strm);
    return 1;	
} 

void ReadSingleColumn(char *path, short x)
{
    FILE *strm;
    
    char s[40];
    
    printf("Opening %s column config file\n",path);
    if ((strm=fopen(path,"r")) == NULL) {
	perror("Can't read config file");
    	if (Tcl_interactive) {
	    fprintf(Tcl_strmw,"Error 1 %s\n",path);
	}
	return;
    }
    fgets(s,40,strm);
    if (strcmp(s,"ExEf column\n")) {
	fprintf(stderr,"Config file broken\n"); 
    	if (Tcl_interactive) {
	    fprintf(Tcl_strmw,"Error 3 %s\n",path);
	}
	fclose(strm);
	return;
    }

    fgets(s,40,strm);
    if (strncmp(s,"*Column",7) == 0) {
	printf("Loading Column %d:%s",x,s+7);
	if (Tcl_interactive) {
	    fprintf(Tcl_strmw,"Column %d%s",x,s+7);
	}
	
	ReadColumn(strm,x);
    }
}

void dropRealtime()
{
    const struct sched_param Schedule={0};

    if (sched_setscheduler(0,SCHED_OTHER,&Schedule) == -1) {
	perror("Warning: priority not dropped");
    }
}
 

/*Initialization functions - Audio*/

void realtime()			/* Setting up realtime mode */
{
    const struct sched_param Schedule={99};	/* setting priority */

    if (mlockall(MCL_CURRENT | MCL_FUTURE) == -1) {
	perror("Warning: can't lock memory");
    }
    if (sched_setscheduler(0,SCHED_FIFO,&Schedule) == -1) {
	perror("Warning: not exclusive realtime mode");
    }
}

void opencard()			/* Opening device */
{
    int spd;
    int format;

    printf("Opening %s\n", Audio_DEVICE_NAME);
    if ((Audio_fd = open(Audio_DEVICE_NAME,O_RDWR|O_NONBLOCK,0)) == -1) {
	perror("Can't open soundcard");
	Errors(10, 0, Audio_DEVICE_NAME);
    }

    printf("Setting blocking mode\n");
    if (fcntl(Audio_fd, F_SETFL, O_RDWR) == -1) {
	perror("Can't set blocking mode");
	Errors(11, 0, Audio_DEVICE_NAME);
    }

    if (!(output_dest || input_src)) {
	printf("Setting full duplex\n");
	if (ioctl(Audio_fd, SNDCTL_DSP_SETDUPLEX, 0) == -1) {
	    perror("Can't try full duplex");
	    Errors(11, 0, Audio_DEVICE_NAME);
	}
	if (ioctl(Audio_fd, SNDCTL_DSP_GETCAPS, &Audio_caps) == -1) {
	    perror("Can't check card's capabilities");
	    Errors(11, 0, Audio_DEVICE_NAME);
	}
	if (!(Audio_caps & DSP_CAP_DUPLEX)) {
	    fprintf(stderr,"Full duplex not supported\n");
	    Errors(11, 0, Audio_DEVICE_NAME);
	}
    }

    if (ioctl(Audio_fd, SNDCTL_DSP_SETFRAGMENT, &Audio_arg) == -1) {
	perror("Can't set up buffers");
	Errors(11, 0, Audio_DEVICE_NAME);
    }

    Audio_format = (Audio_24) ? AFMT_S32_LE: AFMT_S16_LE;
    format = Audio_format;
    if (ioctl(Audio_fd, SNDCTL_DSP_SETFMT, &Audio_format) == -1) {
	perror("Unknown error");
	Errors(11, 0, Audio_DEVICE_NAME);
    }
    if (Audio_24) {
	if (Audio_format != format) {
    	    Audio_format = AFMT_S16_LE;
    	    format = Audio_format;
    	    if (ioctl(Audio_fd, SNDCTL_DSP_SETFMT, &Audio_format) == -1) {
	        perror("Unknown error");
	        Errors(11, 0, Audio_DEVICE_NAME);
    	    }
	    
	    if (Audio_format != AFMT_S16_LE) {
		fprintf(stderr,"No format supported by soundcard\n");
		Errors(11, 0, Audio_DEVICE_NAME);
	    }
	    
	    fprintf(stderr,"Warning: 24 bit format not supported by soundcard\n");
	    Audio_24 = 0;
	    if (Tcl_interactive) {
		fprintf(Tcl_strmw,"Error 4 %s\n", Audio_DEVICE_NAME);
	    }
	} else
	    printf("24 bit mode\n");
    } else
	if (Audio_format != format) {
	    fprintf(stderr,"16 bit format not supported by soundcard\n");
	    Errors(11, 0, Audio_DEVICE_NAME);
	}

    if (ioctl(Audio_fd, SNDCTL_DSP_STEREO, &Audio_stereo) == -1) {
	perror("Unknown error");
	Errors(11, 0, Audio_DEVICE_NAME);
    }
    if (Audio_stereo == 1) {
	fprintf(stderr,"Can't switch to mono mode\n");
	Errors(11, 0, Audio_DEVICE_NAME);
    }

    spd = Audio_speed;
    if (ioctl(Audio_fd, SNDCTL_DSP_SPEED, &Audio_speed) == -1) {
	perror("Unknown error");
	Errors(11, 0, Audio_DEVICE_NAME);
    }
    if (abs(Audio_speed - spd) > 1000) {
	fprintf(stderr,"Warning: sample rate not supported\n");
    	if (Tcl_interactive) {
	    fprintf(Tcl_strmw,"Error 9 %s\n", Audio_DEVICE_NAME);
	}
    }
    printf("Soundcard successfully init'd at %d Hz\n",Audio_speed);
}


/*Initializing funcions - Tcl*/

void IniTmp()
{
    printf("Changing to /var/tmp/.ExEf\n");
    if (chdir("/var/tmp/.ExEf")) {
	perror("Can't change");
	printf("Creating /var/tmp/.ExEf\n");
	if (mkdir("/var/tmp/.ExEf",0777)) {
	    perror("Can't create");
	    Errors(0, 1, NULL);
	} else chmod("/var/tmp/.ExEf",0777);
	if (chdir("/var/tmp/.ExEf")) {
	    perror("Can't change");
	    Errors(0, 1, NULL);
	}
    }
}

void OpenTcl() 
{
    printf("Creating FIFO pipes\n");
    if (mkfifo("FToTcl",0666) == -1) {
	perror("Warning: Can't create forward FIFO pipe");
    } else chmod("FToTcl",0666);
    if (mkfifo("FFromTcl",0666) == -1) {
	perror("Warning: Can't create backward FIFO pipe");
    } else chmod("FFromTcl",0666);
    Tcl_strmw = fopen("FToTcl","w");
    setlinebuf(Tcl_strmw);
    Tcl_filer = open("FFromTcl",O_RDONLY);
    if (flock(Tcl_filer, LOCK_EX | LOCK_NB) == -1) {
	perror("Can't lock pipe");
	Errors(32, 1, NULL);
    }
}


/*DSP Functions*/

void fft(float *real, float *imag)
{
    int delta,ones,br = 0;
    int step,i,j,start;
    float tempr,tempi,WNc,WNs,WNkc,WNks,w;

    br=0;
    for (i=1;i<Audio_BufSize;i++) {
        delta=i^(i-1);
        ones=Audio_BufSize>>1;
	while ((delta>>1) > 0) {
            delta=delta>>1;
	    ones=ones|ones>>1;
    	}
        br = br^ones;
	if (i>br) {
            tempr = real[i]; real[i] = real[br]; real[br] = tempr;
	    tempi = imag[i]; imag[i] = imag[br]; imag[br] = tempi;
    	}
    }

    w=-2*M_PI;
    for (step = 1; step < Audio_BufSize; step = step<<1) {
        w = w * 0.5;
	for (start = 0; start < Audio_BufSize; start += step<<1) {
            WNkc = 1;
	    WNks = 0;
    	    WNc = cosf(w);
            WNs = sinf(w);
	    for (i = start; (i - start) < step; i++) {
        	j = step + i;
                tempr = WNkc*real[j] - WNks*imag[j];
	        tempi = WNks*real[j] + WNkc*imag[j];
    	        real[j] = real[i] - tempr;
        	imag[j] = imag[i] - tempi;
                real[i] += tempr;
	        imag[i] += tempi;
    	        tempr = WNc*WNkc - WNs*WNks;
        	WNks = WNs*WNkc + WNc*WNks;
                WNkc = tempr;
	    }
    	}
    }
}

float Wave(char ftion, float arg)
{
    float temp,res;
    if (ftion == 's') return(sinf(arg));
    else {
	temp = fmodf(arg,1);
	if (temp < 0.5) res = (2*temp - 1);
	else res = (1 - 2*temp);
	return(res);
    }
}


void FGain(short x, short y) {
    short n;
    float g;
    TGain = (Uarray [y] [x]).LinkToBox;
    g = TGain->g;
    for (n = 0; n < Audio_BufSize; n++)	{
	xn[x][n] = g*xn[x][n];
    }
}

void FPass(short x, short y) {
    short n;
    float b0,b1,a1,d0,d1;
    TPass = (Uarray [y] [x]).LinkToBox;
    b0 = TPass->b0;
    b1 = TPass->b1;
    a1 = TPass->a1;
    d1 = TPass->d1;

    for (n = 0; n < Audio_BufSize; n++)	{
	d0 = xn[x][n] - a1*d1;
	xn[x][n] = b0*d0 + b1*d1;
	d1 = d0;
    }
    TPass->d1 = d1;
}

void FBiquad(short x, short y)
{
    short n;
    float b0,b1,b2,a1,a2,d0,d1,d2;
    TBiquad = (Uarray [y] [x]).LinkToBox;

    b0 = TBiquad->b0;
    b1 = TBiquad->b1;
    b2 = TBiquad->b2;
    a1 = TBiquad->a1;
    a2 = TBiquad->a2;
    d1 = TBiquad->d1;
    d2 = TBiquad->d2;
    
    for (n = 0; n < Audio_BufSize; n++)	{
	d0 = xn[x][n] - a1*d1 - a2*d2;
	xn[x][n] = b0*d0 + b1*d1 + b2*d2;
	d2 = d1;
	d1 = d0;
    }
    TBiquad->d2 = d2;
    TBiquad->d1 = d1;
}

void FDistort(short x, short y)
{
    short n;
    char ftion,lftion,rftion;
    float pd, ir = 0;
    TDistort = (Uarray [y] [x]).LinkToBox;
    lftion = TDistort->lftion;
    rftion = TDistort->rftion;
    pd = TDistort->pd;

    for (n = 0; n < Audio_BufSize; n++) {
	ftion = (xn[x][n] < 0) ? lftion:rftion;
	switch(ftion) {
	    case 'a':
		ir = atanf(1.572*xn[x][n])*0.636;
		break;
	    case 's':
		ir = 0.9998*sinf(xn[x][n]);
		break;
	    case 'l':
		ir = 1.9996/(1+expf(-2*xn[x][n])) - 0.9998;
		break;
	    case 'S':
		ir = xn[x][n];
		if (ir < -0.9998) ir = -0.9998;
		if (ir > 0.9998) ir = 0.9998;
		break;
	    case 'z':
		ir = xn[x][n];
		if (ir < -0.9998) ir = 0;
		if (ir > 0.9998) ir = 0;
		break;
	    case 't':
		ir = xn[x][n];
		ir = fmodf(ir,0.9998);
		break;
	    case '1':
		ir = xn[x][n];
		break;
	    case '-':
		ir = -xn[x][n];
		break;
	    case '0':
		ir = 0;
		break;
	    case 'r':
		ir = (xn[x][n] < 0) ? -sqrtf(-xn[x][n] + 0.25) +
		    0.50:sqrtf(xn[x][n] + 0.25) - 0.5;
		break;
	    case 'L':
		ir = (xn[x][n] < 0) ? -logf(1 - xn[x][n]):logf(1 + xn[x][n]);
		break;
	}	
	xn[x][n] = pd*xn[x][n] + ir;
    }
}    

void FHalf(short x, short y)
{
    short n;
    float pd,w1,w2;
    float xnh[2 * Audio_BufSize];
    THalf = (Uarray [y] [x]).LinkToBox;
    pd = THalf->pd;
    for (n = 0; n < Audio_BufSize; n++) {
	xnh[2*n] = xn[x][n];
	xnh[2*n+1] = xn[x][n];
    }
    for (n = 0; n < Audio_BufSize; n++) {
	w1 = .5*(1 - cosf(M_PI*n/(Audio_BufSize)));
	w2 = .5*(1 - cosf(M_PI*(n+Audio_BufSize)/(Audio_BufSize)));
	xn[x][n] = xn[x][n]*pd + w1*xnh[n] + w2*THalf->prev[n];
    }
    for (n = 0; n < Audio_BufSize; n++)	{
	THalf->prev[n] = xnh[n+Audio_BufSize];
	}
    }

void FModulator(short x, short y)
{
    char wave;
    short n = 0,period,interp;
    float pd,delay,depth,arg,prev,DoubleBuf[Audio_BufSize<<1],T,f,wT,b0,d0,d1;
    TModulator = (Uarray [y] [x]).LinkToBox;
    wave = TModulator->wave;
    depth = 0.01*TModulator->depth;
    period = TModulator->period;
    prev = TModulator->PrevGen;
    if (wave == 's')
	arg = 2000*M_PI/(float)(period*Audio_speed);
    else
	arg = 1000/(float)(period*Audio_speed);
    pd = TModulator->pd;
    switch (TModulator->ftion) {
	case 'a':
            for (n = 0; n < Audio_BufSize; n++) {
		xn[x][n] = (1 + depth*Wave(wave,arg*n + prev))*xn[x][n];
	    }
	    break;
	case 'f':
	    for (n = 0; n < Audio_BufSize; n++)	{
		DoubleBuf[n] = TModulator->prev[n];
		DoubleBuf[n + Audio_BufSize] = xn[x][n];
		TModulator->prev[n] = xn[x][n];
	    }
	    for (n = 0; n < Audio_BufSize; n++)	{
		delay = (Audio_BufSize>>1)*depth*Wave(wave,arg*n + prev);
		interp = (Audio_BufSize>>1) + n + (short)floorf(delay);
		xn[x][n] = DoubleBuf[interp] + (DoubleBuf[interp+1]
		- DoubleBuf[interp])*(delay - interp + (Audio_BufSize>>1) + n)
		+ pd*xn[x][n];
	    }
	    break;
	case 'p':
	    d1 = TModulator->d1;
	    T = 1/(float)(Audio_speed);
            for (n = 0; n < Audio_BufSize; n++) {
		f = 1 + 1.5*(depth + depth*Wave(wave,arg*n + prev));
		f = 2 * powf(10,f);
		wT = 2*tanf(M_PI*f*T);
		b0 = (2 - wT)/(2 + wT);
		d0 = xn[x][n] + b0*d1;
		xn[x][n] = - b0*d0 + d1 + pd*xn[x][n];
		d1 = d0;
	    }
	    TModulator->d1 = d1;
    }
    if (wave == 's')
	prev = fmodf (arg*n + prev, 2*M_PI);
    else 
	prev = fmodf (arg*n + prev, 1);
    TModulator->PrevGen = prev;
}	

void FDelay(short x, short y)
{
    unsigned short i;
    short n;
    unsigned short pre,d1,d2;
    float output,R1,R2;
    float B0,B1,B2;
    TDelay = (Uarray [y] [x]).LinkToBox;
    i = TDelay->i;
    pre = TDelay->pre;
    d1 = TDelay->d1;
    d2 = TDelay->d2;
    output = TDelay->output;
    R1 = TDelay->R1;
    R2 = TDelay->R2;

    if (Switch_reset)
	if (TDelay->reset)
	    if(Switch_channel == x)
		Reset(TDelay->D);

    for (n = 0; n < Audio_BufSize; n++)	{
	TDelay->D[i] = xn[x][n] - R1*R2*TDelay->D[(unsigned short)(i-d1-d2)];
	B1 = 0.5*R1*TDelay->D[(unsigned short)(i-pre-d1)];
	B2 = 0.5*R2*TDelay->D[(unsigned short)(i-pre-d2)];
	B0 = R1*R2*TDelay->D[(unsigned short)(i-pre-d1-d2)];
	xn[x][n] = xn[x][n] + (B1 + B2 + B0)*output;
	i++;
    }
    TDelay->i = i;
}

void FHall(short x, short y)
{
    unsigned short i;
    short n;
    unsigned short pre,d1,d2,d3;
    float output,R1,R2,R3;
    THall = (Uarray [y] [x]).LinkToBox;
    i = THall->i;
    pre = THall->pre;
    d1 = THall->d1;
    d2 = THall->d2;   
    d3 = THall->d3;
    output = THall->output;
    R1 = THall->R1;
    R2 = THall->R2;
    R3 = THall->R3;

    if (Switch_reset)
	if (THall->reset)
	    if(Switch_channel == x)
		Reset(THall->D);

    for (n = 0; n < Audio_BufSize; n++)	{
	THall->D[i] = xn[x][n] - R1*THall->D[(unsigned short)(i-d1)]
	    - R2*THall->D[(unsigned short)(i-d2)]
	    - R3*THall->D[(unsigned short)(i-d3)];
	xn[x][n] = xn[x][n] + output*(R1*THall->D[(unsigned short)(i-d1-pre)]
	    + R2*THall->D[(unsigned short)(i-d2-pre)]
	    + R3*THall->D[(unsigned short)(i-d3-pre)]);
	i++;
    }
    THall->i = i;
}

void FReverb(short x, short y)
{
    unsigned short i;
    short n,k,ord;
    unsigned short d,fbtime;
    float fbk,b,sumx,output;
    TReverb = (Uarray [y] [x]).LinkToBox;
    i = TReverb->i;
    ord = TReverb->ord;
    fbtime = TReverb->fbtime;
    fbk = TReverb->fbk;
    output = TReverb->output;

    if (Switch_reset)
	if (TReverb->reset)
	    if(Switch_channel == x)
		Reset(TReverb->D);

    for (n = 0; n < Audio_BufSize; n++) {
	sumx = xn[x][n];
        TReverb->D[i] = sumx - fbk*TReverb->D[(unsigned short)(i-fbtime)];
	for (k = 0; k < ord; k++) {
	    d = TReverb->Bi[k];
	    b = TReverb->Bv[k];
	    sumx = sumx + b*TReverb->D[(unsigned short)(i-d)];    
	}
	xn[x][n] = xn[x][n] + sumx*output;
	i++;
    }
    TReverb->i = i;
}

void FCompressor(short x, short y)
{
    short n;
    float lambda,thresh,ratio,clambda,dlambda,power,g,threshp,inp,ing;
    TCompressor = (Uarray [y] [x]).LinkToBox;
    lambda = TCompressor->lambda;
    thresh = TCompressor->thresh;
    ratio = TCompressor->ratio;
    clambda = TCompressor->clambda;
    dlambda = TCompressor->dlambda;
    power = TCompressor->powerp;
    g = TCompressor->gp;
    threshp = thresh*thresh;
    for (n = 0; n < Audio_BufSize; n++) {
	inp = xn[x][n]*xn[x][n];
        power = power*lambda + (1-lambda)*inp;
    	ing = (power<threshp) ? 1 : (powf(power/threshp,0.5*ratio - 0.5));
	g = (ing<g) ? g*clambda + (1-clambda)*ing : g*dlambda + (1-dlambda)*ing;
	xn[x][n] = g*xn[x][n];
    }
    TCompressor->powerp = power;
    TCompressor->gp = g;
}

void FQuant(short x, short y)
{
    short n;
    float levels;
    TQuant = (Uarray [y] [x]).LinkToBox;
    levels = TQuant->levels;
    for (n = 0; n < Audio_BufSize; n++)	{
	xn[x][n] = rintf(xn[x][n]*levels) / levels;
    }
}

void FNoiseGate(short x, short y)
{
    short n;
    float lambda,threshp,saturp,clambda,dlambda,power,g,inp;
    TNoiseGate = (Uarray [y] [x]).LinkToBox;
    lambda = TNoiseGate->lambda;
    threshp = TNoiseGate->threshp;
    saturp = TNoiseGate->saturp;
    clambda = TNoiseGate->clambda;
    dlambda = TNoiseGate->dlambda;
    power = TNoiseGate->powerp;
    g = TNoiseGate->gp;
    for (n = 0; n < Audio_BufSize; n++) {
	inp = xn[x][n]*xn[x][n];
        power = power*lambda + (1-lambda)*inp;
        if (power>saturp*threshp) power = saturp*threshp;
    	if (power>threshp) {
	    g = g*clambda + (1-clambda);
	}
	else g = g*dlambda;
        xn[x][n] = g*xn[x][n];
    }

    TNoiseGate->powerp = power;
    TNoiseGate->gp = g;
}

void FSpecSubF(short x, short y)
{
    short n,limit;
    float time,w;
    float xseg1[Audio_BufSize],xseg2[Audio_BufSize];
    float power = 0,thresh,lambda,ratio;
    float Sn[Audio_BufSize];
    float real[Audio_BufSize],imag[Audio_BufSize];
    TSpecSubF = (Uarray [y] [x]).LinkToBox;
    thresh = TSpecSubF->thresh;
    lambda = TSpecSubF->lambda;
    time = TSpecSubF->time;
    limit = TSpecSubF->limit;
    w = 2*M_PI/Audio_BufSize;

    for (n = 0; n < (Audio_BufSize>>1); n++) {
	xseg2[n] = TSpecSubF->PrevClean[n] * (1 + cosf(w*n));
	real[n] = TSpecSubF->PrevSig[n];
	real[n + (Audio_BufSize>>1)] = xn[x][n];
	TSpecSubF->PrevSig[n] = xn [x] [n + (Audio_BufSize>>1)];
	imag[n] = 0;
	imag[n + (Audio_BufSize>>1)] = 0;
    }
    
    fft(real,imag);

    if (limit < 0) {
	time = -1e6;
    }

    if (time < limit) {
        for (n = 0; n<Audio_BufSize; n++) {
    	    power = power + xn[x][n]*xn[x][n];
	}
        power = power/Audio_BufSize;

	if (power<thresh) {
	    for (n = 0; n<Audio_BufSize; n++) {
	        TSpecSubF->Noise[n] = TSpecSubF->Noise[n]*lambda + (1-lambda) * hypotf(real[n],imag[n]);
	    }
	}	    
        time = time + 1000*Audio_BufSize/(float) Audio_speed;
    }


    for (n = 0; n<Audio_BufSize; n++) {
        Sn[n] = copysignf(hypotf(real[n],imag[n]) - TSpecSubF->Noise[n], real[n]);
	if (real[n] == 0) {
	    imag[n] = -Sn[n];
	} else {
	    ratio = imag[n] / real[n];
	    real[n] = Sn[n] / sqrt(1 + ratio*ratio);
	    imag[n] = -real[n] * ratio;
	}	
    }	

    fft(real,imag);
    
    for (n = 0; n<Audio_BufSize; n++) {
	xseg1[n] = real[n] / Audio_BufSize * (1 - cosf(w*n));
	real[n] = xn[x][n];
	imag[n] = 0;
    }

    fft(real,imag);	
    
    for (n = 0; n<Audio_BufSize; n++) {
        Sn[n] = copysignf(hypotf(real[n],imag[n]) - TSpecSubF->Noise[n], real[n]);
	if (real[n] == 0) {
	    imag[n] = -Sn[n];
	} else {
	    ratio = imag[n] / real[n];
	    real[n] = Sn[n] / sqrt(1 + ratio*ratio);
	    imag[n] = -real[n] * ratio;
	}	
    }

    fft(real,imag);

    for (n = 0; n<(Audio_BufSize>>1); n++) {
	xseg2[n + (Audio_BufSize>>1)] = real[n] / Audio_BufSize * (1 - cosf(w*n));
	TSpecSubF->PrevClean[n] = real[n + (Audio_BufSize>>1)] / Audio_BufSize;
    }

    for (n = 0; n < (Audio_BufSize); n++) {
	xn[x][n] = (xseg1[n] +xseg2[n]) * 0.5;
    }
 
    TSpecSubF->time = time;
}

void FSinGen(short x, short y)		/*Sinus -20dB*/
{
    short n,freq;
    float prev,pd,w;
    TSinGen = (Uarray [y] [x]).LinkToBox;
    prev = TSinGen->prev;
    pd = TSinGen->pd;
    freq = TSinGen->freq;
    w = 2*M_PI*(float)freq/Audio_speed;
    for (n = 0; n < Audio_BufSize; n++)	{
	xn[x][n] = xn[x][n]*pd + 0.1*sinf(w*n + prev);
    }
    prev = fmodf (w*n + prev, 2*M_PI);	
    TSinGen->prev = prev;
}
    
void FWNoise(short x, short y)		/*White Noise -20dB*/
{
    short n;
    float pd,r;
    TWNoise = (Uarray [y] [x]).LinkToBox;
    pd = TWNoise->pd;
    for (n = 0; n < Audio_BufSize; n++) {
	r = ((float) rand())/RAND_MAX;
	xn[x][n] = pd*xn[x][n] + 0.039*logf(1/r-1);
    }
}

void FCopyToNextColumn(short x)
{
    short n;
    for (n = 0; n < Audio_BufSize; n++)
	xn [x + 1] [n] = xn[x][n];
}

void FInterVU(short x)
{
    if (Tcl_interactive) {
	EstPower(xn[x],&VU_interpower,&VU_over,VU_lambda);
        VU_interover = VU_interover | VU_over;
	VU_inpower = VU_interpower;
	VU_inover = VU_interover;
    }
}

void SelChan()
{
    short n;
    for (n = 0; n < Audio_BufSize; n++) {
	Switch_gain = Switch_lambda*Switch_gain + (1 - Switch_lambda);
	xno [n] = Switch_gain * xn [Switch_channel] [n] + (1 - Switch_gain) * xn [Switch_prevchan] [n];
    }
}

void EstPower(float *sig, float *power, int *over, float lambda) {
    int ov = 0;
    short n;
    float inp,estp;
    estp = *power;
    for (n = 0; n < Audio_BufSize; n++) {
	inp = sig[n]*sig[n];
	estp = estp*lambda + (1-lambda)*inp;
	if (fabsf(sig[n])>0.9999) ov = 1;
    }
    *power = estp;
    *over = ov;
}

void EvalCommand()
{
    short x,y;
    char a;
    char s[40],path[80];
    
    printf(Tcl_line);
    if (!strncmp(Tcl_line,"Box",3)) {
	sscanf(Tcl_line+4,"%hd %hd %c",&x,&y,&a);
	LoadBox(Tcl_line+10,x,y,a);
    }
    if (!strncmp(Tcl_line,"Switch",6)) {
	sscanf(Tcl_line+7,"%hd %hd",&x,&y);
	if (Switch_channel != x) {
	    Switch_prevchan = Switch_channel;
	    Switch_channel = x;
	    Switch_gain = 1 - Switch_gain;
	    Switch_lambda = Forget(y);
	    Switch_reset = 1;
	}
    }
    if (!strncmp(Tcl_line,"VUMeters",8)) {
	sscanf(Tcl_line + 9,"%hd %hd %d",&x,&VU_refresh,&VU_onoff);
	VU_lambda = Forget(x);
    }
    if (!strncmp(Tcl_line,"Audio",5)) {	
	close(Audio_fd);
	sscanf(Tcl_line + 6,"%d %hd %hd",&Audio_speed,&x,&y);
	Audio_arg = 0x00040000 + x;
	Audio_BufSize = 1<<(x-1);
	Audio_24 = (y>16) ? 1:0;
	opencard();
	checkout();
        printf("fragstotal %d\n",Audio_info.fragstotal);
	printf("fragsize %d B\n",Audio_info.fragsize);
        Audio_maxfrag = Audio_info.fragstotal;
    }
    if (!strncmp(Tcl_line,"Device",6)) {
	sscanf(Tcl_line + 7,"%s",Audio_DEVICE_NAME);
    }
    if (!strncmp(Tcl_line,"SaveWave",8)) {
	sscanf(Tcl_line + 9,"%d %hd %s %s",&Save_saving,&x,Save_name,s);
	if (!strcmp(s,"input"))
	    Save_point = 1;
	else Save_point = 2;
	Save_maxfrags = 1<<x;
    }
    if (!strncmp(Tcl_line,"ChangeBox",9)) {
	sscanf(Tcl_line+10,"%hd %hd %c",&x,&y,&a);
	ChangeBox(x,y,a);
    }
    if (!strncmp(Tcl_line,"RemoveBox",9)) {
	sscanf(Tcl_line+10,"%hd %hd",&x,&y);
	RemoveBox(x,y);
    }
    if (!strncmp(Tcl_line,"InsertBox",9)) {
	sscanf(Tcl_line+10,"%hd %hd",&x,&y);
	InsertBox(x,y);
    }
    if (!strncmp(Tcl_line,"Load",4)) {
	sscanf(Tcl_line+5,"%s",path);
	close(Audio_fd);
	RemoveAll();
	ReadConfig(path);
	opencard();
	checkout();
        printf("fragstotal %d\n",Audio_info.fragstotal);
	printf("fragsize %d B\n",Audio_info.fragsize);
        Audio_maxfrag = Audio_info.fragstotal;
    }
    if (!strncmp(Tcl_line,"NewMatrix",9)) {
	RemoveAll();
    }
    if (!strncmp(Tcl_line,"Column",6)) {
	sscanf(Tcl_line+7,"%hd %s",&x,path);
	RemoveColumn(x);
	ReadSingleColumn(path,x);
    }
    if (!strncmp(Tcl_line,"WaveLoad",8)) {
    	sscanf(Tcl_line+9,"%s",path);
	if ((input_fd = open(path,O_RDONLY)) == -1) {
	    perror("Can't open file");
	    input_fd = Audio_fd;
	    fprintf(Tcl_strmw,"Error 1 %s\n",path);
	} else {
	    close(Audio_fd);
	    input_src = 1;
	    dropRealtime();
	    opencard();
	    checkout();
            printf("fragstotal %d\n",Audio_info.fragstotal);
	    printf("fragsize %d B\n",Audio_info.fragsize);
            Audio_maxfrag = Audio_info.fragstotal;
	}
    }
}

void ReadCommand()
{
    ssize_t readStatus;
    char receiv[1];
    fd_set rfds;
    struct timeval tv;
    FD_ZERO(&rfds);
    FD_SET(Tcl_filer,&rfds);
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    while (select(Tcl_filer + 1,&rfds,NULL,NULL,&tv)) {
	readStatus = read(Tcl_filer,receiv,1);
	if (readStatus > 0) {
	    if (receiv[0] != 0) {
		Tcl_line[Tcl_charpos] = receiv[0];
        	Tcl_charpos++;
	    }
	    if (receiv[0] == 10) {
		Tcl_line[Tcl_charpos] = 0;
        	Tcl_charpos=0;
		EvalCommand();
	    }
	}
	if (!readStatus) {
	    exit(0);
	}
	if (readStatus == -1) {
	    if (Tcl_interactive) {
		fprintf(Tcl_strmw,"Error 31 X");
	    }
	}
    }
}
    
void effect()
{
    short n,x,y;
    float dB;

    for (x=0; x < 4; x++)
	if (Audio_24)					 /* buffer -> x[n] */
	    for (n = 0; n < Audio_BufSize; n++)		   
		xn[x][n]=(float) Audio_buffer_32[n]/2147483648.0;
	else
            for (n = 0; n < Audio_BufSize; n++)
		xn[x][n]=(float) Audio_buffer[n]/32768.0;
	    
    VU_nth--;
    if (Tcl_interactive) {
	ReadCommand();
	EstPower(xn[0],&VU_inpower,&VU_over,VU_lambda);
        VU_inover = VU_inover | VU_over;
    }

    for (x=0; x < 4; x++) {
        y=0;
	while((Uarray [y] [x]).LinkToBox != NULL) {
	    switch((Uarray [y] [x]).KindOfBox) {
		case 'g':
		    FGain(x,y);
		    break;
	        case 'b':
		    FBiquad(x,y);
		    break;
		case 'p':
		    FPass(x,y);
		    break;
		case 'd':
		    FDistort(x,y);
		    break;
		case 'h':
		    FHalf(x,y);
		    break;
		case 'm':
		    FModulator(x,y);
		    break;
		case 'D':
		    FDelay(x,y);
		    break;
		case 'H':
		    FHall(x,y);
		    break;
		case 'R':
		    FReverb(x,y);
		    break;
		case 'C':
		    FCompressor(x,y);
		    break;
		case 'q':
		    FQuant(x,y);
		    break;
		case 'n':
		    FNoiseGate(x,y);
		    break;
		case 'S':
		    FSpecSubF(x,y);
		    break;
		case 's':
		    FSinGen(x,y);
		    break;
		case 'w':
		    FWNoise(x,y);
		    break;
		case 'c':
		    FCopyToNextColumn(x);
		    break;
		case 'v':
		    FInterVU(x);
		    break;
	    }
	    y++;
	}
    }
    
    Switch_reset = 0;
    
    SelChan();

    for (n = 0; n < Audio_BufSize; n++) {
	if (xno [n] <= -1) xno [n] = -1;
	if (xno [n] >= 1) xno [n] = 1;
    }
	
    if (Audio_24)					/* x[n] -> buffer */
	for (n = 0; n < Audio_BufSize; n++)
	    Audio_buffer_32[n]=(int)rintf(xno [n]*2147483647.0);
    else
        for (n = 0; n < Audio_BufSize; n++)
	    Audio_buffer[n]=(short)rintf(xno [n]*32767.0);

    if (Tcl_interactive) {
	EstPower(xno,&VU_outpower,&VU_over,VU_lambda);
	VU_outover = VU_outover | VU_over;
	if (!VU_nth) {
	    VU_nth = VU_refresh;
	    if (VU_onoff) {
		dB = (10*log10f(2*VU_inpower));
	        if (dB < -120) dB = -120;
		fprintf(Tcl_strmw,"Input %d %d\n",(short)floorf(dB),VU_inover);
		VU_inover = 0;
		VU_interover = 0;
		dB = (10*log10f(2*VU_outpower));
		if (dB < -120) dB = -120;
		fprintf(Tcl_strmw,"Output %d %d\n",(short)floorf(dB),VU_outover);
		VU_outover = 0;
	    }
	}
    }
}


/*Realtime Audio Functions*/

void SaveWave()
{
    int n,shift;
    int Bps;
    int i = 0;

    if (Audio_24)
	Bps = 4;
    else
	Bps = 2;

    switch (Save_saving) {
	case 1:
	    Save_saving++;
	    do {
		start = malloc(16 + Bps*Audio_BufSize*Save_maxfrags);
		i++;
	    } while ((start == NULL) && (i < 3));
	    if (start == NULL) {
		fprintf(stderr, "Memory allocation failed, can't start recording\n");
		Save_saving--;
		if (Tcl_interactive) {
		    fprintf(Tcl_strmw,"Error 22 X\n");
		}
		break;
	    }
	    current = start;
	    Save_fragwpage = 0;
	    Save_pages = 1;
	case 2:
	    if (Save_fragwpage >= Save_maxfrags) {
		Save_fragwpage = 0;
		do {
		    current->next = malloc(16 + Bps*Audio_BufSize*Save_maxfrags);
		    i++;
		} while ((current->next == NULL) && (i < 3));
		if (current->next == NULL) {
		    fprintf(stderr, "Memory allocation failed, continuing\n");
		} else {
		    current = current->next;
		    Save_pages++;
		}
	    }

	    shift = Save_fragwpage*Audio_BufSize;
	    if (Audio_24)
		for (n=0;n<Audio_BufSize;n++)
		    current->data_32[n+shift] = Audio_buffer_32[n];
	    else
		for (n=0;n<Audio_BufSize;n++)
		    current->data[n+shift] = Audio_buffer[n];

	    Save_fragwpage++;
	    break;
	case 4:
	    current->next = NULL;
	    SaveToDisk();
	    Save_saving = 0;
	    break;
    }
}

void checkout()
{
    if (ioctl(Audio_fd, SNDCTL_DSP_GETOSPACE, &Audio_info) == -1) {
    	perror("Output error");
    }
}

void writefrag()
{
    int len;
    if (output_dest == 0) {
	if (Audio_24) {
	    if ((len = write(Audio_fd, Audio_buffer_32, Audio_BufSize<<2)) == -1) {
		perror("Can't play");
    	    }
	} else {
	    if ((len = write(Audio_fd, Audio_buffer, Audio_BufSize<<1)) == -1) {
		perror("Can't play");
    	    }
	}
    }

    if ((Save_point == 2) && Save_saving)
	SaveWave();
}

int readfrag()
{
    int len;
    if (Audio_24) {
	if ((len = read(input_fd, Audio_buffer_32, Audio_BufSize<<2)) == -1) {
	    perror("Can't record");
    	}
    } else {
	if ((len = read(input_fd, Audio_buffer, Audio_BufSize<<1)) == -1) {
	    perror("Can't record");
    	}
    }

    if ((Save_point == 1) && Save_saving)
        SaveWave();
    return len;
}

void play()
{
    int len;
    
    do {
	len = readfrag();
	if (len == 0)
	    InEof();
	if (output_dest == 0) {
	    checkout();
	    if (Audio_info.fragments == Audio_maxfrag)
	    writefrag();
	}
	effect();
        writefrag();
    } while (!0);
}


int main(int argc, char **argv)
{
    char *home;
    char defaultFile[80];
    char *file, *infile = NULL;
    file = defaultFile;

    home = getenv("HOME");
    if (home != NULL) {
    	sprintf(defaultFile, "%s/%s", home, ".ExEf/.ExEfrc");
    }

    printf("\nExEf - the ultimate effect engine\nCopyright (C) 2001-2003 Tom Olexa\n\n");
    
    CmdLine(argc, argv, &file, &infile);
    
    if (infile) {
	printf("Input from %s file\n", infile);
	if ((input_fd = open(infile,O_RDONLY)) == -1) {
	    perror("Can't open file");
	    if (Tcl_interactive) {
		fprintf(Tcl_strmw,"Error 1 %s\n", infile);
	    }    
	}
	else
	    input_src = 1;
    }

    if (file != defaultFile)
	if (!ReadConfig(file)) file = defaultFile;
    if (file == defaultFile) {
    	HomeDir("/.ExEf");
        if (!ReadConfig(file)) {
	    printf("Copying config file from /usr/share/ExEf\n");
	    system("cp -v -f /usr/share/ExEf/.ExEfrc ~/.ExEf");
	    ReadConfig(".ExEfrc");
	}
    }

    if (!(output_dest || input_src))
	realtime();

    if (!(output_dest && input_src)) {
	opencard();

	checkout();
	printf("fragments %d\n",Audio_info.fragments);
	printf("fragstotal %d\n",Audio_info.fragstotal);
	printf("fragsize %d B\n",Audio_info.fragsize);
	printf("bytes %d\n",Audio_info.bytes);
	Audio_maxfrag = Audio_info.fragstotal;
    }

    if (!input_src)
	input_fd = Audio_fd;

    atexit(SaveToDisk);
    signal(SIGHUP,Signals);
    signal(SIGINT,Signals);
    signal(SIGPIPE,Signals);
    signal(SIGTERM,Signals);
    
    play();
    close(Audio_fd);

    return(0);
}
